﻿using System;
using Utilities.Math;

namespace GpsShield
{
    public class GprmcParser
    {
        // Parse the GPRMC line
        //
        public static GpsPoint Parse(string line)
        {
            // $GPRMC,040302.663,A,3939.7,N,10506.6,W,0.27,358.86,200804,,*1A

            if(!IsCheckSumGood(line))
            {
                return null;
            }

            try
            {
                string[] parts = line.Split(',');
                if (parts.Length != 12)
                {
                    return null;
                }

                if (parts[2] != "A")
                {
                    return null;
                }

                string date = parts[9]; // UTC Date DDMMYY
                if (date.Length != 6)
                {
                    return null;
                }
                int year = 2000 + int.Parse(date.Substring(4, 2));
                int month = int.Parse(date.Substring(2, 2));
                int day = int.Parse(date.Substring(0, 2));
                string time = parts[1]; // HHMMSS.XXX
                if (time.Length != 10)
                {
                    return null;
                }
                int hour = int.Parse(time.Substring(0, 2));
                int minute = int.Parse(time.Substring(2, 2));
                int second = int.Parse(time.Substring(4, 2));
                int milliseconds = int.Parse(time.Substring(7, 3));
                DateTime utcTime = new DateTime(year, month, day, hour, minute, second, milliseconds);

                string lat = parts[3];  // HHMM.MMMM
                if (lat.Length != 9)
                {
                    return null;
                }
                double latHours = double.Parse(lat.Substring(0, 2));
                double latMins = double.Parse(lat.Substring(2));
                double latitude = latHours + latMins / 60.0;
                if (parts[4] == "S")       // N or S
                {
                    latitude = -latitude;
                }

                string lng = parts[5];  // HHHMM.M
                if (lng.Length != 10)
                {
                    return null;
                }
                double lngHours = double.Parse(lng.Substring(0, 3));
                double lngMins = double.Parse(lng.Substring(3));
                double longitude = lngHours + lngMins / 60.0;
                if (parts[6] == "W")
                {
                    longitude = -longitude;
                }

            	double speed = 0.0;
				if(parts[7]!=string.Empty)
				{
					speed = double.Parse(parts[7]);
				}
            	double bearing = 0.0;
				if(parts[8]!=string.Empty)
				{
					bearing = double.Parse(parts[8]);
				}

                // Should probably validate check sum

                GpsPoint gpsPoint = new GpsPoint
                                      {
                                          BearingInDegrees = bearing,
                                          Latitude = latitude,
                                          Longitude = longitude,
                                          SpeedInKnots = speed,
                                          Timestamp = utcTime
                                      };
                return gpsPoint;
                
            }
            catch (Exception)
            {
                // One of our parses failed...ignore.
            }
            return null;
        }

        private static bool IsCheckSumGood(string sentence)
        {
            int index1 = sentence.IndexOf("$");
            int index2 = sentence.LastIndexOf("*");

            if (index1 != 0 || index2 != sentence.Length - 3 )
            {
                return false;
            }

            string checkSumString = sentence.Substring(index2 + 1, 2);
            int checkSum1 = Convert.ToInt32(checkSumString, 16);

            string valToCheck = sentence.Substring(index1 + 1, index2 - 1);
            char c = valToCheck[0];
            for(int i = 1;i<valToCheck.Length;i++)
            {
                c ^= valToCheck[i];
            }

            return checkSum1 == c;
        }
    }
}