﻿using System;
using System.IO.Ports;
using System.Threading;
using Microsoft.SPOT;

namespace GpsShield
{
    public class Reader
    {
        private readonly object _lock = new object();
        private readonly SerialPort _serialPort;
        private readonly int _timeOut;
        private readonly double _minDistanceBetweenPoints;
        private bool _isStarted;
        private Thread _processor;

        public delegate void LineProcessor(string line);

        public delegate void GpsDataProcessor(GpsPoint gpsPoint, string rawLine);

        public event LineProcessor RawLine;
        public event GpsDataProcessor GpsData;

        public bool IsStarted { get { return _isStarted; } }

        public Reader(SerialPort serialPort)
            : this(serialPort, 100, 0.0)
        {

        }
        public Reader(SerialPort serialPort, int timeOutBetweenReadsInMilliseconds, double minDistanceInMilesBetweenPoints)
        {
            _serialPort = serialPort;
            _timeOut = timeOutBetweenReadsInMilliseconds;
            _minDistanceBetweenPoints = minDistanceInMilesBetweenPoints;
        }

        public bool Start()
        {
            lock (_lock)
            {
                if(_isStarted)
                {
                    return false;
                }
                _isStarted = true;
                _processor = new Thread(ThreadProc);
                _processor.Start();
            }
            return true;
        }

        public bool Stop()
        {
            lock (_lock)
            {
                if(!_isStarted)
                {
                    return false;
                }
                _isStarted = false;
                if (!_processor.Join(5000))
                {
                    _processor.Abort();
                }
                return true;
            }
        }

        private void ThreadProc()
        {
            Debug.Print("GPS thread started...");
            if(!_serialPort.IsOpen)
            {
                _serialPort.Open();
            }
            while (_isStarted)
            {
                int bytesToRead = _serialPort.BytesToRead;
                if (bytesToRead > 0)
                {
                    byte[] buffer = new byte[bytesToRead];
                    int bytesRead = _serialPort.Read(buffer, 0, buffer.Length);
                    try
                    {
                        string temp = GetAllValidChars(buffer, bytesRead);
                        //string temp = new string(System.Text.Encoding.UTF8.GetChars(buffer));
                        ProcessBytes(temp);
                    }
                    catch (Exception ex)
                    {
                        // only process lines we can parse.
                        char[] chars = new char[buffer.Length];
                        for(int i=0;i<buffer.Length;i++)
                        {
                            chars[i] = (char) buffer[i];
                        }
                        Debug.Print(new string(chars));
                        Debug.Print(ex.ToString());
                    }
                }

                Thread.Sleep(_timeOut);
            }
            Debug.Print("GPS thread stopped...");
        }

        private static string GetAllValidChars(byte[] buffer, int bytesToConvert)
        {
            const byte upper = 128;

            char[] tmpret = new char[bytesToConvert];
            int curLoc = 0;
            for (int i = 0; i < bytesToConvert; i++)
            {
                if (buffer[i] < upper)
                {
                    tmpret[curLoc++] = (char)buffer[i];
                }
            }

            return new string(tmpret, 0, curLoc);
        }

        private string _data = string.Empty;
        private GpsPoint _lastPoint;
        private void ProcessBytes(string temp)
        {
			if(temp == string.Empty)
			{
				return;
			}
            string[] parts = temp.Split('\n');
            int numToProcess = parts.Length - 1;
            if (temp[temp.Length - 1] == '\n')
            {
            	numToProcess = parts.Length;
            }
			for (int i = 0; i < numToProcess;i++ )
			{
				_data += parts[i];
				_data = _data.Trim();
				if (_data == string.Empty)
				{
					continue;
				}
				if (_data.IndexOf("$GPRMC") == 0)
				{
					if (GpsData != null)
					{
						GpsPoint gpsPoint = GprmcParser.Parse(_data);
						if (gpsPoint != null)
						{
							bool isOk = true;
							if (_lastPoint != null)
							{
								double distance = GeoDistanceCalculator.DistanceInMiles(gpsPoint.Latitude, gpsPoint.Longitude,
								                                                        _lastPoint.Latitude, _lastPoint.Longitude);
								double distInFeet = distance * 5280;
								Debug.Print("distance = " + distance + " mi (" + distInFeet + " feet)");
								if (distance < _minDistanceBetweenPoints)
								{
									// Too close to the last point....don't raise the event
									isOk = false;
									Debug.Print("Point is too close: distance=" + distance + ", min=" +
									            _minDistanceBetweenPoints);
								}
							}
							_lastPoint = gpsPoint;

							// Raise the event
							if (isOk)
							{
								GpsData(gpsPoint, _data);
							}
						}
					}
				}
				if (RawLine != null)
				{
					RawLine(_data);
				}
			}
			if (numToProcess != parts.Length)
			{
				// There is left over data.
				_data = parts[parts.Length - 1];
			}
			else
			{
				_data = string.Empty;
			}
        }
    }
}